import os
from math import ceil
from pathlib import Path
from typing import Optional

import pandas as pd
import streamlit as st

from helix.components.configuration import plot_options_box
from helix.components.images.logos import sidebar_logo
from helix.options.choices.ui import PROBLEM_TYPES
from helix.options.data import DataOptions
from helix.options.enums import ExecutionStateKeys, PlotOptionKeys, ProblemTypes
from helix.options.execution import ExecutionOptions
from helix.options.file_paths import helix_experiments_base_dir, uploaded_file_path
from helix.options.plotting import PlottingOptions
from helix.services.data import read_uploaded_data
from helix.services.experiments import create_experiment
from helix.utils.utils import save_upload


def _directory_is_valid(directory: Path) -> bool:
    """Determine if the directory supplied by the user is valid. If it already exists,
    it is invalid.

    Args:
        directory (Path): The path to check.

    Returns:
        bool: `True` if the directory doesn't already exist, else `False`
    """
    return not directory.exists()


def _save_directory_selector() -> Path:
    """Create a selector for the directory to save experiments."""
    root = helix_experiments_base_dir()

    col1, col2 = st.columns(2, vertical_alignment="bottom")

    col1.text(f"{root}{os.path.sep}", help="Your experiment will be saved here")
    sub_dir = col2.text_input("Name of the experiment", placeholder="e.g. MyExperiment")

    return root / sub_dir


def _file_is_uploaded() -> bool:
    """Determine if the user has uploaded a file to the form.

    Returns:
        bool: `True` if a file was uploaded, else `False`.
    """
    return st.session_state.get(ExecutionStateKeys.UploadedFileName) is not None


def infer_problem_type_from_column(target: pd.Series) -> Optional[ProblemTypes]:
    """
    Infer the problem type given the target column. If the column contains categorical
    values (e.g. strings or few unique integers), return Classification. Otherwise, Regression.

    Returns:
        Optional[ProblemTypes]: The inferred problem type, or None if error occurs.
    """

    # Heuristic: if target is string/object or has few unique integers → classification
    if target.dtype == "object" or target.dtype.name == "category":
        return ProblemTypes.Classification
    elif pd.api.types.is_integer_dtype(target):
        n_unique = target.nunique()
        if (
            n_unique < 20 or n_unique <= len(target) * 0.05
        ):  # we assume that less than 20 unique values or less than 5% of the data are unique values indicates a classification problem
            return ProblemTypes.Classification
        else:
            return ProblemTypes.Regression
    elif pd.api.types.is_float_dtype(target):
        return ProblemTypes.Regression
    else:
        return None


def _entrypoint(save_dir: Path):
    """Function to serve as the entrypoint for experiment creation, with access
    to the session state. This is so configuration captured in fragements is
    passed correctly to the services in this function.

    Args:
        save_dir (Path): The path to the experiment.
    """
    # Set up options to save
    path_to_data = uploaded_file_path(
        st.session_state[ExecutionStateKeys.UploadedFileName].name,
        helix_experiments_base_dir()
        / st.session_state[ExecutionStateKeys.ExperimentName],
    )
    exec_opts = ExecutionOptions(
        problem_type=st.session_state.get(
            ExecutionStateKeys.ProblemType, ProblemTypes.Auto
        ).lower(),
        random_state=st.session_state[ExecutionStateKeys.RandomSeed],
        dependent_variable=st.session_state[ExecutionStateKeys.DependentVariableName],
        experiment_name=st.session_state[ExecutionStateKeys.ExperimentName],
    )
    data_opts = DataOptions(
        data_path=str(path_to_data),  # Path objects aren't JSON serialisable
        target_column=st.session_state[ExecutionStateKeys.DependentVariable],
        feature_columns=st.session_state[ExecutionStateKeys.FeatureColumns],
        # use `.get` as IdColumn is optional and can be None
        id_column=st.session_state.get(ExecutionStateKeys.IdColumn),
    )
    plot_opts = PlottingOptions(
        plot_axis_font_size=st.session_state[PlotOptionKeys.AxisFontSize],
        plot_axis_tick_size=st.session_state[PlotOptionKeys.AxisTickSize],
        plot_colour_scheme=st.session_state[PlotOptionKeys.ColourScheme],
        plot_colour_map=st.session_state[PlotOptionKeys.ColourMap],
        angle_rotate_xaxis_labels=st.session_state[PlotOptionKeys.RotateXAxisLabels],
        angle_rotate_yaxis_labels=st.session_state[PlotOptionKeys.RotateYAxisLabels],
        save_plots=st.session_state[PlotOptionKeys.SavePlots],
        plot_title_font_size=st.session_state[PlotOptionKeys.TitleFontSize],
        plot_font_family=st.session_state[PlotOptionKeys.FontFamily],
        dpi=st.session_state[PlotOptionKeys.DPI],
        height=st.session_state[PlotOptionKeys.Height],
        width=st.session_state[PlotOptionKeys.Width],
    )

    # Create the experiment directory and save configs
    create_experiment(
        save_dir,
        plotting_options=plot_opts,
        execution_options=exec_opts,
        data_options=data_opts,
    )

    # Save the data
    uploaded_file = st.session_state[ExecutionStateKeys.UploadedFileName]
    save_upload(path_to_data, uploaded_file, data_opts)


st.set_page_config(
    page_title="New Experiment",
    page_icon=sidebar_logo(),
)

st.header("New Experiment")
st.write(
    """
    Here you can create a new experiment. Once created, you will be able
    to select it on the Data Preprocessing, Data Visualisation and Train Models pages.
    """
)
st.write(
    """
    ### Create a new experiment ⚗️

    Give your experiment a name, upload your data, and click **Create**, located at the bottom of the page.
    If an experiment with the same name already exists, or you don't provide a file,
    you will not be able to create it.
    """
)

save_dir = _save_directory_selector()
# If a user has tried to enter a destination to save an experiment, show it
# if it's valid, else show some red text showing the destination and saying
# it's invalid.
is_valid = _directory_is_valid(save_dir)
if not is_valid and st.session_state.get(ExecutionStateKeys.ExperimentName):
    st.warning(
        f"Cannot use {save_dir}; it already exists. If you have just created this experiment, please continue."
    )
else:
    st.session_state[ExecutionStateKeys.ExperimentName] = (
        save_dir.name
    )  # get the experiment name from the file path

st.write(
    """
    Upload your data as a CSV or Excel (.xslx) file.
    """
)

st.write(
    """
    **Please note that the last column of the uploaded file should be the dependent variable.**
    """
)

uploaded_file = st.file_uploader(
    "Choose a file",
    type=["csv", "xlsx"],
    key=ExecutionStateKeys.UploadedFileName,
    help="Updload a CSV or Excel (.xslx) file containing your data.",
)


def select_target_column(dataframe: pd.DataFrame) -> str:
    """ """
    columns = dataframe.columns.tolist()

    target_col = st.selectbox(
        "Select the target column (dependent variable) to model",
        options=columns,
        index=len(columns) - 1,  # Default to last column
        help="Select the column you want to predict. By default, the last column is selected.",
        key=ExecutionStateKeys.DependentVariable,
    )

    return target_col


suggested_problem_type = None

if uploaded_file is not None:

    uploaded_file.seek(0)
    data = read_uploaded_data(uploaded_file=uploaded_file)
    if st.checkbox("Show uploaded data"):
        st.markdown(" #### Uploaded data")
        st.dataframe(data)

    target_col = select_target_column(data)

    suggested_problem_type = infer_problem_type_from_column(data[target_col])
    st.write(
        f"**Suggested problem type based on the selected column `{target_col}`: {suggested_problem_type}**"
    )

    st.text_input(
        "Name of the dependent variable. **This will be used for the plots. As default, the name of the selected column of the uploaded file will be used**",
        value=target_col,
        key=ExecutionStateKeys.DependentVariableName,
    )

    feature_cols = [col for col in data.columns if col != target_col]

    # TODO: allow user to select ID columns

    n_features = len(feature_cols)
    group_size = 150

    if st.toggle(
        "Select the feature columns manually (optional)",
        help="By default, all columns except the target column will be used as features. You can manually select specific ones below.",
    ):
        # Initialize checkbox states if not set
        if "feature_checkboxes" not in st.session_state:
            st.session_state.feature_checkboxes = {col: True for col in feature_cols}

        st.markdown("#### Select which features to include:")

        # Split features into chunks of group_size
        num_groups = ceil(n_features / group_size)
        for i in range(num_groups):
            start_idx = i * group_size
            end_idx = min((i + 1) * group_size, n_features)
            group = feature_cols[start_idx:end_idx]
            first_feat, last_feat = group[0], group[-1]

            with st.expander(
                f"Features {start_idx + 1}–{end_idx} ({len(group)}) | {first_feat} → {last_feat}"
            ):
                # Select-all toggle for this group
                select_all_key = f"group_select_{i}"

                select_all = st.toggle(
                    "Select all features in this group",
                    value=False,
                    key=select_all_key,
                )

                st.divider()
                cols = st.columns(5)  # Adjust layout columns per row

                for j, col_name in enumerate(group):
                    with cols[j % 5]:
                        st.session_state.feature_checkboxes[col_name] = st.checkbox(
                            col_name,
                            value=(select_all),
                            key=f"chk_{col_name}",
                        )

        # Extract selected features
        selected_features = [
            col
            for col, checked in st.session_state.feature_checkboxes.items()
            if checked
        ]

        # Save to session_state
        st.session_state[ExecutionStateKeys.FeatureColumns] = selected_features

        # Display summary
        st.write(f"✅ **Total selected features: `{len(selected_features)}`**")
        if not selected_features:
            st.error(
                "No features selected! Please select some features to create an experiment."
            )
            is_valid = False

    else:
        # Default: all columns except target
        st.session_state[ExecutionStateKeys.FeatureColumns] = [
            col for col in data.columns if col != target_col
        ]
        st.write(f"Using all {len(feature_cols)} feature columns by default.")

    if st.toggle(
        "Select ID column",
        help="If you have a column containing row IDs, check this box to select it.",
    ):
        id_cols = list(
            filter(
                lambda x: not (
                    x in st.session_state[ExecutionStateKeys.FeatureColumns]
                    or x == target_col
                ),
                data.columns,
            )
        )
        st.selectbox(
            "ID column",
            options=id_cols,
            placeholder="Select...",
            index=None,
            key=ExecutionStateKeys.IdColumn,
        )

st.selectbox(
    "Problem type",
    PROBLEM_TYPES,
    key=ExecutionStateKeys.ProblemType,
    index=(
        PROBLEM_TYPES.index(suggested_problem_type.capitalize())
        if suggested_problem_type
        else 1
    ),
)
st.info(
    """
    If your dependent variable is categorical (e.g. cat 🐱 or dog 🐶), choose **"Classification"**.

    If your dependent variable is continuous (e.g. stock prices 📈), choose **"Regression"**.
    """
)

st.number_input(
    "Random seed",
    value=1221,
    min_value=0,
    key=ExecutionStateKeys.RandomSeed,
    help="Setting this allows experiments to be repeatable despite random shuffling of data.",
)

# Set up plotting options for the experiment
st.subheader("Configure experiment plots")
plot_options_box()

if st.button(
    "Create",
    type="primary",
    disabled=not is_valid or not _file_is_uploaded(),
    on_click=_entrypoint,
    args=(save_dir,),
):
    st.success(f"Experiment created at `{save_dir}`")



import logging
import os
import shutil
import subprocess
from pathlib import Path

from helix.utils.utils import create_directory


def main():
    """The entrypoint of Helix.
    This method shouldn't be called explicitly. Use the `helix` command in the
    terminal after installing the app.
    """
    app_path = os.path.join(os.path.dirname(__file__), "Home.py")
    source_config = Path(os.path.dirname(__file__)) / ".streamlit" / "config.toml"
    dest_config = Path.home() / ".streamlit"
    create_directory(dest_config)
    shutil.copy(source_config, dest_config)
    try:
        subprocess.run(["streamlit", "run", app_path])
    except KeyboardInterrupt:
        logging.info("Shutting down Helix...")



from dataclasses import dataclass

from helix.options.enums import DataSplitMethods, Normalisations


@dataclass
class DataSplitOptions:
    """
    Options class specifying how to split data in bootstrapping or cross-validation,
    as well as the size of the test size.
    """

    method: DataSplitMethods
    n_bootstraps: int | None = None
    k_folds: int | None = None
    test_size: float = 0.2


@dataclass
class DataOptions:
    """
    Options class specifying where data are saved, how to split them, and which
    normalisation technique to use.
    """

    data_path: str | None = None
    target_column: str | None = None
    feature_columns: list[str] | None = None
    data_split: DataSplitOptions | None = None
    normalisation: Normalisations = Normalisations.NoNormalisation
    id_column: str | None = None



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import shap

from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.file_paths import (
    fi_options_dir,
    fi_plot_dir,
    fi_result_dir,
    fuzzy_plot_dir,
    fuzzy_result_dir,
    helix_experiments_base_dir,
)
from helix.options.plotting import PlottingOptions
from helix.utils.logging_utils import Logger
from helix.utils.utils import log_options


def save_importance_results(
    feature_importance_df: pd.DataFrame,
    model_type,
    importance_type: str,
    feature_importance_type: str,
    experiment_name: str,
    fi_opt: FeatureImportanceOptions,
    plot_opt: PlottingOptions,
    logger: Logger,
    shap_values=None,
):
    """Save the feature importance results to a CSV file and the plots.

    Args:
        feature_importance_df (pd.DataFrame): DataFrame of feature importance results.
        model_type (_type_): Type of model.
        importance_type (str): Type of feature importance method.
        feature_importance_type (str): Type of feature importance method (Again for some reason).
        experiment_name (str): Name of the experiment, to know where to save outputs.
        fi_opt (FeatureImportanceOptions): Feature importance options.
        plot_opt (PlottingOptions): Plotting options.
        logger (Logger): The logger.
        shap_values (_type_, optional): SHAP values. Defaults to None.
    """

    # Return early if feature_importance_df is empty
    if feature_importance_df.empty:
        logger.info(f"No importance results for {feature_importance_type}...")
        return

    biofefi_base_dir = helix_experiments_base_dir()
    logger.info(f"Saving importance results and plots of {feature_importance_type}...")

    # Save plots when the flag is set to True and importance type is not fuzzy
    if fi_opt.save_feature_importance_plots and importance_type != "fuzzy":
        save_dir = fi_plot_dir(biofefi_base_dir / experiment_name)
        if not save_dir.exists():
            save_dir.mkdir(exist_ok=True, parents=True)
        # Plot bar plot - sort values in descending order and plot top n features
        # rotate x-axis labels for better readability
        plt.style.use(plot_opt.plot_colour_scheme)
        fig, ax = plt.subplots(layout="constrained", dpi=plot_opt.dpi)

        top_features = (
            feature_importance_df.sort_values(by=0, ascending=False)
            .head(fi_opt.num_features_to_plot)
            .T
        )
        sns.barplot(top_features, ax=ax)
        # rotate x-axis labels for better readability
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=plot_opt.angle_rotate_xaxis_labels,
            family=plot_opt.plot_font_family,
        )
        ax.set_yticklabels(
            ax.get_yticklabels(),
            rotation=plot_opt.angle_rotate_yaxis_labels,
            family=plot_opt.plot_font_family,
        )
        ax.set_xlabel("Feature", family=plot_opt.plot_font_family)
        ax.set_ylabel("Importance", family=plot_opt.plot_font_family)
        ax.set_title(
            f"{feature_importance_type} - {model_type}",
            family=plot_opt.plot_font_family,
            wrap=True,
        )
        fig.savefig(save_dir / f"{model_type}-bar.png")

        if feature_importance_type == "SHAP":
            # Plot bee swarm plot
            fig, ax = plt.subplots(layout="constrained", dpi=plot_opt.dpi)
            ax.set_title(
                f"{feature_importance_type} - {model_type}",
                family=plot_opt.plot_font_family,
                wrap=True,
            )
            shap.plots.beeswarm(
                shap_values, max_display=fi_opt.num_features_to_plot, show=False
            )
            ax.set_xlabel(ax.get_xlabel(), family=plot_opt.plot_font_family)
            ax.set_ylabel(ax.get_ylabel(), family=plot_opt.plot_font_family)
            ax.set_xticklabels(
                ax.get_xticklabels(),
                family=plot_opt.plot_font_family,
            )
            ax.set_yticklabels(
                ax.get_yticklabels(),
                family=plot_opt.plot_font_family,
            )
            fig.savefig(save_dir / f"{model_type}-beeswarm.png")

    # Save the results to a CSV file - create folders if they don't exist
    if fi_opt.save_feature_importance_results and importance_type != "fuzzy":
        save_dir = fi_result_dir(biofefi_base_dir / experiment_name)
        if not save_dir.exists():
            save_dir.mkdir(exist_ok=True, parents=True)
        feature_importance_df.to_csv(save_dir / f"{feature_importance_type}.csv")

    if fi_opt.save_feature_importance_results and importance_type == "fuzzy":
        save_dir = fuzzy_result_dir(biofefi_base_dir / experiment_name)
        if not save_dir.exists():
            save_dir.mkdir(exist_ok=True, parents=True)
        feature_importance_df.to_csv(save_dir / f"{feature_importance_type}.csv")

    # Save the metrics to a log file
    if fi_opt.save_feature_importance_options:
        options_path = fi_options_dir(biofefi_base_dir / experiment_name)
        if not options_path.exists():
            options_path.mkdir(parents=True, exist_ok=True)
        log_options(options_path, fi_opt)


def save_fuzzy_sets_plots(
    universe,
    membership_functions,
    x_cols,
    exec_opt: ExecutionOptions,
    plot_opt: PlottingOptions,
    logger: Logger,
):
    # Plot the membership functions
    logger.info("Saving fuzzy set plots ...")
    save_dir = fuzzy_plot_dir(helix_experiments_base_dir() / exec_opt.experiment_name)
    if not save_dir.exists():
        save_dir.mkdir(exist_ok=True, parents=True)

    plt.style.use(plot_opt.plot_colour_scheme)
    plt.rcParams["font.family"] = plot_opt.plot_font_family
    for feature in x_cols:
        fig, ax = plt.subplots(layout="constrained", dpi=plot_opt.dpi)
        ax.plot(
            universe[feature],
            membership_functions[feature]["small"],
            "r",
            label="Small",
        )
        ax.plot(
            universe[feature],
            membership_functions[feature]["moderate"],
            "g",
            label="Moderate",
        )
        ax.plot(
            universe[feature],
            membership_functions[feature]["large"],
            "b",
            label="Large",
        )
        ax.set_title(
            f"{feature} Membership Functions",
            family=plot_opt.plot_font_family,
            wrap=True,
        )
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=plot_opt.angle_rotate_xaxis_labels,
            family=plot_opt.plot_font_family,
        )
        ax.set_yticklabels(
            ax.get_yticklabels(),
            rotation=plot_opt.angle_rotate_yaxis_labels,
            family=plot_opt.plot_font_family,
        )
        ax.legend()
        fig.savefig(save_dir / f"fuzzy set-{feature}.png".replace("/", ""))
    plt.close()


def save_target_clusters_plots(
    df_cluster, exec_opt: ExecutionOptions, plot_opt: PlottingOptions, logger: Logger
):
    # Plot the target clusters
    logger.info("Saving target clusters plot ...")
    save_dir = fuzzy_plot_dir(helix_experiments_base_dir() / exec_opt.experiment_name)
    if not save_dir.exists():
        save_dir.mkdir(exist_ok=True, parents=True)

    # Plot boxplot of the target (continuous values) and target clusters (categories) using seaborn
    plt.style.use(plot_opt.plot_colour_scheme)
    fig, ax = plt.subplots(layout="constrained", dpi=plot_opt.dpi)
    sns.boxplot(data=df_cluster, x="cluster", y="target", hue="cluster", ax=ax)
    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=plot_opt.angle_rotate_xaxis_labels,
        family=plot_opt.plot_font_family,
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        rotation=plot_opt.angle_rotate_yaxis_labels,
        family=plot_opt.plot_font_family,
    )
    ax.set_title(
        "Target Clusters",
        family=plot_opt.plot_font_family,
        wrap=True,
    )
    fig.savefig(save_dir / "fuzzy-target clusters.png")
    plt.close()



"""Train Models page for Helix.

This page allows users to configure and train machine learning models on their data.
"""

from pathlib import Path

import pandas as pd
import streamlit as st

from helix.components.configuration import display_options
from helix.components.experiments import experiment_selector
from helix.components.forms.forms_ml_opts import ml_options_form
from helix.components.images.logos import sidebar_logo
from helix.components.logs import log_box
from helix.components.plots import display_metrics_table, display_predictions, plot_box
from helix.machine_learning import train
from helix.options.data import DataOptions
from helix.options.enums import (
    DataSplitMethods,
    ExecutionStateKeys,
    MachineLearningStateKeys,
    PlotOptionKeys,
)
from helix.options.execution import ExecutionOptions
from helix.options.file_paths import (
    data_options_path,
    execution_options_path,
    helix_experiments_base_dir,
    log_dir,
    ml_metrics_full_path,
    ml_metrics_mean_std_path,
    ml_model_dir,
    ml_options_path,
    ml_plot_dir,
    ml_predictions_path,
    plot_options_path,
)
from helix.options.ml import MachineLearningOptions
from helix.options.plotting import PlottingOptions
from helix.services.configuration import (
    load_data_options,
    load_execution_options,
    load_plot_options,
    save_options,
)
from helix.services.data import TabularData, ingest_data
from helix.services.experiments import get_experiments
from helix.services.logs import get_logs
from helix.services.ml_models import (
    models_exist,
    save_model,
    save_model_predictions,
    save_models_metrics,
)
from helix.utils.logging_utils import Logger, close_logger
from helix.utils.utils import delete_directory, set_seed


def build_configuration() -> (
    tuple[MachineLearningOptions, ExecutionOptions, PlottingOptions, DataOptions, str]
):
    """Build the configuration options to run the Machine Learning pipeline.

    Returns:
        tuple[MachineLearningOptions, ExecutionOptions, PlottingOptions, DataOptions, str]:
        The machine learning options, general execution options, plotting options,
        data options, experiment name.
    """

    experiment_name = st.session_state[ExecutionStateKeys.ExperimentName]

    path_to_plot_opts = plot_options_path(
        helix_experiments_base_dir() / experiment_name
    )
    plot_opt = load_plot_options(path_to_plot_opts)

    path_to_exec_opts = execution_options_path(
        helix_experiments_base_dir() / experiment_name
    )
    path_to_data_opts = data_options_path(
        helix_experiments_base_dir() / experiment_name
    )

    exec_opt = load_execution_options(path_to_exec_opts)
    ml_opt = MachineLearningOptions(
        save_actual_pred_plots=st.session_state[PlotOptionKeys.SavePlots],
        model_types=st.session_state[MachineLearningStateKeys.ModelTypes],
        ml_plot_dir=str(ml_plot_dir(helix_experiments_base_dir() / experiment_name)),
        ml_log_dir=str(log_dir(helix_experiments_base_dir() / experiment_name) / "ml"),
        save_models=st.session_state[MachineLearningStateKeys.SaveModels],
        use_hyperparam_search=st.session_state.get(
            ExecutionStateKeys.UseHyperParamSearch, True
        ),
    )
    data_opts = load_data_options(path_to_data_opts)
    # update data opts
    data_opts.data_split = st.session_state.get(ExecutionStateKeys.DataSplit)
    save_options(path_to_data_opts, data_opts)

    return ml_opt, exec_opt, plot_opt, data_opts, experiment_name


def pipeline(
    ml_opts: MachineLearningOptions,
    exec_opts: ExecutionOptions,
    plotting_opts: PlottingOptions,
    data_opts: DataOptions,
    experiment_name: str,
    data: TabularData,
):
    """This function actually performs the steps of the pipeline. It can be wrapped
    in a process it doesn't block the UI.

    Args:
        ml_opts (MachineLearningOptions): Options for machine learning.
        exec_opts (ExecutionOptions): General execution options.
        plotting_opts (PlottingOptions): Options for plotting.
        experiment_name (str): The name of the experiment.
        data (TabularData): The data that will be used in the pipeline.
    """
    seed = exec_opts.random_state
    set_seed(seed)
    logger_instance = Logger(Path(ml_opts.ml_log_dir))
    logger = logger_instance.make_logger()
    n_cpus = st.session_state[ExecutionStateKeys.NCPUs]

    # Machine learning
    trained_models, metrics_full, metrics_mean_std = train.run(
        ml_opts=ml_opts,
        data_opts=data_opts,
        plot_opts=plotting_opts,
        data=data,
        exec_opts=exec_opts,
        logger=logger,
        n_cpus=n_cpus,
    )
    if ml_opts.save_models:
        predictions = pd.DataFrame(
            columns=["Y True", "Y Prediction", "Model Name", "Set", "Bootstrap"]
        )

        for model_name in trained_models:

            for i, model in enumerate(trained_models[model_name]):
                save_path = (
                    ml_model_dir(helix_experiments_base_dir() / experiment_name)
                    / f"{model_name}-{i}.pkl"
                )
                save_model(model, save_path)

                predictions_train = model.predict(data.X_train[i])
                predictions_train = {
                    "Y True": data.y_train[i],
                    "Y Prediction": predictions_train,
                    "Model Name": model_name,
                    "Set": "Train",
                    "Bootstrap": i,
                }
                df_train = pd.DataFrame(predictions_train)
                predictions_test = model.predict(data.X_test[i])
                predictions_test = {
                    "Y True": data.y_test[i],
                    "Y Prediction": predictions_test,
                    "Model Name": model_name,
                    "Set": "Test",
                    "Bootstrap": i,
                }
                df_test = pd.DataFrame(predictions_test)
                df_whole = pd.concat([df_train, df_test], ignore_index=True)
                if data.id_column is not None:
                    df_whole.insert(
                        loc=0, column=data.id_column[i].name, value=data.id_column[i]
                    )
                predictions = pd.concat([predictions, df_whole], ignore_index=True)
        st.session_state[MachineLearningStateKeys.Predictions] = predictions

    if ml_opts.use_hyperparam_search:
        predictions = predictions[["Y True", "Y Prediction", "Model Name", "Set"]]
    elif data_opts.data_split.method == DataSplitMethods.KFold:
        predictions = predictions.rename(columns={"Bootstrap": "Fold"})

    # Save full metrics
    save_models_metrics(
        metrics_full,
        ml_metrics_full_path(helix_experiments_base_dir() / experiment_name),
    )
    # Save mean/std metrics
    save_models_metrics(
        metrics_mean_std,
        ml_metrics_mean_std_path(helix_experiments_base_dir() / experiment_name),
    )
    save_model_predictions(
        predictions,
        ml_predictions_path(helix_experiments_base_dir() / experiment_name),
    )
    # Close the logger
    close_logger(logger_instance, logger)


st.set_page_config(
    page_title="Train Models",
    page_icon=sidebar_logo(),
)
sidebar_logo()

st.header("Train Models")
st.write(
    """
    In this page you can train new machine learning models. For that:

    > 1. Select an experiment.
    >
    > 2. Choose whether or not you want Helix to perform hyperparameter search.
    >
    > 3. Specify how you wish the data to be split into training and test data.
    >
    > 4. Select the models you wish to train (you may choose more than one).
    >
    > 5. Choose which outputs to save and hit **"Run Training"**.
    >
    > 6. Wait for the pipeline to finish. This might take a little while.
    """
)

choices = get_experiments()
experiment_name = experiment_selector(choices)
if experiment_name:
    st.session_state[ExecutionStateKeys.ExperimentName] = experiment_name
    helix_base_dir = helix_experiments_base_dir()
    experiment_dir = helix_base_dir / experiment_name
    display_options(experiment_dir)
    path_to_exec_opts = execution_options_path(experiment_dir)
    exec_opt = load_execution_options(path_to_exec_opts)

    already_trained_models = models_exist(
        ml_model_dir(
            helix_experiments_base_dir()
            / st.session_state[ExecutionStateKeys.ExperimentName]
        )
    )
    if already_trained_models:
        st.warning("⚠️ You have trained models in this experiment.")
        st.checkbox(
            "Would you like to rerun the experiments? This will overwrite the existing models.",
            value=False,
            key=MachineLearningStateKeys.RerunML,
        )
    else:
        st.session_state[MachineLearningStateKeys.RerunML] = False

    if not already_trained_models or st.session_state[MachineLearningStateKeys.RerunML]:
        # Get problem type from execution options
        exec_opts = load_execution_options(
            execution_options_path(helix_experiments_base_dir() / experiment_name)
        )
        st.write("### Configure machine learning models")
        ml_options_form(problem_type=exec_opts.problem_type)
    else:
        st.info(
            "You have chosen not to re-run the machine learning experiments. "
            "You can proceed to feature importance analysis."
        )
        st.stop()

    if st.button("Run Training", type="primary"):

        if experiment_dir.exists():
            delete_directory(ml_model_dir(experiment_dir))
            delete_directory(ml_plot_dir(experiment_dir))
            delete_directory(ml_metrics_mean_std_path(experiment_dir))
            delete_directory(ml_metrics_full_path(experiment_dir))
            delete_directory(ml_predictions_path(experiment_dir))

        ml_opts, exec_opts, plot_opts, data_opts, exp_name = build_configuration()
        # Create the logger
        logger_instance = Logger()
        logger = logger_instance.make_logger()
        # Retrieve data from state or load it from disk
        data = ingest_data(exec_opts, data_opts, logger)
        # Save ML options
        save_options(ml_options_path(experiment_dir), ml_opts)
        with st.spinner("Model training in progress. Check the logs for progress."):
            # wait for the process to finish
            pipeline(ml_opts, exec_opts, plot_opts, data_opts, exp_name, data)
        try:
            st.session_state[MachineLearningStateKeys.MLLogBox] = get_logs(
                log_dir(experiment_dir) / "ml"
            )
            log_box(
                box_title="Machine Learning Logs", key=MachineLearningStateKeys.MLLogBox
            )
        except NotADirectoryError:
            pass

        metrics = ml_metrics_mean_std_path(experiment_dir)
        display_metrics_table(metrics)

        if st.session_state.get(MachineLearningStateKeys.Predictions) is not None:
            display_predictions(
                st.session_state.get(MachineLearningStateKeys.Predictions)
            )
        else:
            predictions = ml_predictions_path(helix_base_dir / experiment_name)
            if predictions.exists():
                preds = pd.read_csv(predictions)
                display_predictions(preds)

        ml_plots = ml_plot_dir(experiment_dir)
        plot_box(ml_plots, "Machine learning plots")



from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LinearRegression,
    LogisticRegression,
    Ridge,
    RidgeClassifier,
)
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from xgboost import XGBClassifier, XGBRegressor

from helix.machine_learning.models.BRNNs import BRNNClassifier, BRNNRegressor
from helix.machine_learning.models.mlrem import EMLinearRegression
from helix.machine_learning.models.random_forest import (
    RandomForestClassifier,
    RandomForestRegressor,
)
from helix.machine_learning.models.svm import SVC, SVR
from helix.options.enums import ModelNames

# Grid search parameters for each model
LINEAR_MODEL_GRID = {
    "fit_intercept": [True, False],
}

RANDOM_FOREST_GRID = {
    "n_estimators": [10, 50, 100, 200],
    "max_depth": [None, 10, 20, 30],
    "min_samples_split": [2, 5, 10],
}

XGB_GRID = {
    "n_estimators": [10, 50, 100, 200],
    "max_depth": [3, 5, 7, 9],
    "learning_rate": [0.01, 0.1, 0.3],
    "subsample": [0.5, 0.8, 1.0],
}

SVM_GRID = {
    "kernel": ["linear", "poly", "rbf"],
    "C": [0.1, 1.0, 10.0],
    "degree": [2, 3, 4],
}

MLREM_GRID = {
    "alpha": [0.05, 0.1, 0.5],
    "max_beta": [40],
    "weight_threshold": [1e-3, 1e-2],
    "max_iterations": [300],
    "tolerance": [0.001],
}

CLASSIFIERS: dict[ModelNames, type] = {
    ModelNames.LinearModel: LogisticRegression,
    ModelNames.RandomForest: RandomForestClassifier,
    ModelNames.XGBoost: XGBClassifier,
    ModelNames.SVM: SVC,
    ModelNames.Ridge: RidgeClassifier,
    ModelNames.KNearestNeighbours: KNeighborsClassifier,
    ModelNames.MLP: MLPClassifier,
    ModelNames.GNB: GaussianNB,
    ModelNames.BRNN: BRNNClassifier,
}

REGRESSORS: dict[ModelNames, type] = {
    ModelNames.LinearModel: LinearRegression,
    ModelNames.RandomForest: RandomForestRegressor,
    ModelNames.XGBoost: XGBRegressor,
    ModelNames.SVM: SVR,
    ModelNames.MLREM: EMLinearRegression,
    ModelNames.Lasso: Lasso,
    ModelNames.ElasticNet: ElasticNet,
    ModelNames.Ridge: Ridge,
    ModelNames.KNearestNeighbours: KNeighborsRegressor,
    ModelNames.MLP: MLPRegressor,
    ModelNames.BRNN: BRNNRegressor,
}



import os
from pathlib import Path


def get_logs(log_dir: Path) -> str:
    """Get the latest log file for the latest run to display.

    Args:
        log_dir (Path): The directory to search for the latest logs.

    Raises:
        NotADirectoryError: `log_dir` does not point to a directory.

    Returns:
        str: The text of the latest log file.
    """
    if not log_dir.is_dir():
        raise NotADirectoryError(f"{log_dir} is not a directory")

    files = list(log_dir.iterdir())
    ctimes = [os.path.getctime(f) for f in files]
    most_recent = max(ctimes)
    index_most_recent = ctimes.index(most_recent)

    with open(files[index_most_recent], "r") as log:
        text = log.read()

    return text



from pathlib import Path

import pandas as pd
import streamlit as st
from scipy.stats import mode
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from helix.components.configuration import display_options
from helix.components.experiments import experiment_selector, model_selector
from helix.components.images.logos import sidebar_logo
from helix.options.enums import (
    ExecutionStateKeys,
    Normalisations,
    PredictStateKeys,
    ProblemTypes,
)
from helix.options.file_paths import (
    data_options_path,
    data_preprocessing_options_path,
    execution_options_path,
    helix_experiments_base_dir,
    ml_model_dir,
)
from helix.options.preprocessing import PreprocessingOptions
from helix.services.configuration import (
    load_data_options,
    load_data_preprocessing_options,
    load_execution_options,
)
from helix.services.data import read_data, read_uploaded_data
from helix.services.experiments import get_experiments
from helix.services.ml_models import load_models
from helix.services.preprocessing import find_non_numeric_columns
from helix.utils.logging_utils import Logger

# Set page contents
st.set_page_config(
    page_title="Predict",
    page_icon=sidebar_logo(),
)


def preprocess_data(
    X: pd.DataFrame,
    predict_data: pd.DataFrame,
    options: PreprocessingOptions | None,
    id_col: str | None = None,
) -> pd.DataFrame:
    if options is None or not options.data_is_preprocessed:
        return predict_data

    columns_to_drop = [id_col] if id_col is not None else []
    columns_to_drop.extend(find_non_numeric_columns(X))
    if columns_to_drop:
        X = X.drop(columns=columns_to_drop)
        predict_data = predict_data.drop(columns=columns_to_drop)

    scaler = get_scaler(options.independent_variable_normalisation)
    if scaler is None:
        return predict_data

    scaler.fit(X)
    return pd.DataFrame(
        scaler.transform(predict_data),
        columns=predict_data.columns,
        index=predict_data.index,
    )


def get_scaler(normalisation_method):
    match normalisation_method:
        case Normalisations.NoNormalisation:
            return None
        case Normalisations.Standardisation:
            return StandardScaler()
        case Normalisations.MinMax:
            return MinMaxScaler()


def ensemble_predictions(
    predictions_df: pd.DataFrame,
    problem_type: ProblemTypes,
) -> pd.Series:
    if problem_type == ProblemTypes.Regression:
        return pd.Series(
            predictions_df.mean(axis=1),
            index=predictions_df.index,
            name="Mean Prediction",
        )

    if problem_type == ProblemTypes.Classification:
        values, _ = mode(predictions_df.values, axis=1, keepdims=False)
        return pd.Series(
            values,
            index=predictions_df.index,
            name="Majority Vote",
        )

    raise ValueError(f"Unsupported problem type: {problem_type}")


def get_model_predictions(models_to_use, trained_models, predict_data):
    predictions = {}

    for model_name, model in trained_models.items():
        if model_name not in models_to_use:
            continue

        display_name = model_name.split(".")[0].replace("-", " ").capitalize()
        predictions[display_name] = model.predict(predict_data)

    return pd.DataFrame(predictions)


def get_predictions(
    raw_data: pd.DataFrame,
    independent_variable_col_names: list,
    predict_data: pd.DataFrame,
    preprocessing_options: PreprocessingOptions | None,
    models: list,
    problem_type: ProblemTypes,
    id_column: str | None = None,
):

    X = raw_data[independent_variable_col_names]
    target_col = (
        predict_data[id_column]
        if id_column and id_column in predict_data.columns
        else predict_data.index.to_series(name=id_column)
    )
    predict_data = predict_data[independent_variable_col_names]

    predict_data = preprocess_data(X, predict_data, preprocessing_options)

    trained_models = load_models(
        ml_model_dir(
            helix_experiments_base_dir()
            / st.session_state[ExecutionStateKeys.ExperimentName]
        )
    )

    predictions_df = get_model_predictions(models, trained_models, predict_data)

    ensemble = ensemble_predictions(predictions_df, problem_type)

    result = pd.concat([predict_data, predictions_df, ensemble], axis=1)
    if target_col is not None:
        result = pd.concat([target_col, result], axis=1)

    st.dataframe(result)


st.header("Predict")
st.write(
    """
    This page allows the prediction of the target variable using the trained models.
    The idea is to use this page to predict the target variable on datasets where this value is unknown.
    """
)


choices = get_experiments()
experiment_name = experiment_selector(choices)
base_dir = helix_experiments_base_dir()
# Create the logger
logger_instance = Logger()
logger = logger_instance.make_logger()


if experiment_name:
    st.session_state[ExecutionStateKeys.ExperimentName] = experiment_name

    display_options(base_dir / experiment_name)

    path_to_exec_opts = execution_options_path(
        helix_experiments_base_dir() / experiment_name
    )
    exec_opt = load_execution_options(path_to_exec_opts)

    path_to_data_options = data_options_path(
        helix_experiments_base_dir() / experiment_name
    )
    data_options = load_data_options(path_to_data_options)

    path_to_preprocessing_options = data_preprocessing_options_path(
        helix_experiments_base_dir() / experiment_name
    )
    if path_to_preprocessing_options.exists():
        preprocessing_options = load_data_preprocessing_options(
            path_to_preprocessing_options
        )
        is_preprocessed = preprocessing_options.data_is_preprocessed
        raw_data_path = data_options.data_path.replace("_preprocessed", "")
    else:
        preprocessing_options = None
        is_preprocessed = False
        raw_data_path = data_options.data_path
    # This is the data that the user provided initially.
    # Having this is useful as it is needed to fit the scalers if needed.
    raw_data = read_data(Path(raw_data_path), logger)

    # This is the data that was actually used for the training (after preprocessing)
    # This is only needed to get the names of the variables used for model training.
    if data_options.feature_columns is not None:
        independent_variables = data_options.feature_columns
    else:
        independent_variables = read_data(Path(data_options.data_path), None).columns[
            :-1
        ]

    st.write(
        "For this experiment, the following columns were used as independent variables:"
    )

    uploaded_file = st.file_uploader(
        "Upload the data ",
        type=["csv", "xlsx"],
        key=PredictStateKeys.PredictFile,
        help="Updload a CSV or Excel (.xslx) file containing your data.",
    )

    if uploaded_file:
        predict_data = read_uploaded_data(uploaded_file)
        st.write("#### View the provided data")
        st.dataframe(predict_data)

    model_dir = ml_model_dir(base_dir / experiment_name)
    if model_dir.exists():
        model_choices = list(
            filter(lambda x: x.endswith(".pkl"), [x.name for x in model_dir.iterdir()])
        )
    else:
        model_choices = []

    if len(model_choices) == 0:
        st.info(
            "You don't have any trained models in this experiment. "
            "Go to **Train Models** to create some models to evaulate."
        )

    else:
        models = model_selector(
            options=model_choices,
            gui_text="Select the models to use for the predictions of your new dataset",
            placeholder="Models for predictions",
            key=PredictStateKeys.PredictModels,
        )

    if uploaded_file:

        missing_cols = []
        for col in independent_variables:
            if col not in predict_data.columns:
                missing_cols.append(col)
                st.error(f"The columns {col} was not found in the provided data.")

        if missing_cols:
            st.error(
                "Dependent variables necessary for the prediciton were not found. Please provide the data with all the variables that you provided in your original data."
            )

        st.success(
            "All the needed independent variables for the predictions were found successfully."
        )

    if st.button(
        "Predict",
        key=PredictStateKeys.PredictButton,
        disabled=not models or not uploaded_file,
    ):
        get_predictions(
            raw_data=raw_data,
            independent_variable_col_names=independent_variables,
            predict_data=predict_data,
            preprocessing_options=preprocessing_options,
            models=models,
            problem_type=exec_opt.problem_type,
            id_column=data_options.id_column,
        )



import os

import streamlit as st

PATH_TO_ICON = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "static/Helix_Logo_Transparent.png",
)


def header_logo():
    """Generate the header logo for the app."""
    _, col2, _ = st.columns(3)

    with col2:
        st.image(PATH_TO_ICON, use_column_width=True)


def sidebar_logo():
    """Generate the sidebar logo in the top left."""
    st.logo(PATH_TO_ICON)



from helix.options.choices.ui import SVM_KERNELS
from helix.options.enums import ActivationFunctions

LINEAR_MODEL_GRID = {
    "fit_intercept": [True, False],
}

LASSO_GRID = {"fit_intercept": [True, False], "alpha": [0.05, 0.1, 0.5, 0.8, 1.0]}

RIDGE_GRID = {"fit_intercept": [True, False], "alpha": [0.05, 0.1, 0.5, 0.8, 1.0]}

ELASTICNET_GRID = {
    "fit_intercept": [True, False],
    "alpha": [0.05, 0.1, 0.5, 0.8, 1.0],
    "l1_ratio": [0, 0.5, 1],
}

KNN_GRID = {"n_neighbors": [5, 10, 15], "leaf_size": [10, 15, 30, 45, 60], "p": [1, 2]}

GNB_GRID = {"var_smoothing": [1e-09, 1e-08]}

MLP_GRID = {
    "hidden_layer_sizes": [(25,), (50,), (75,)],
    "activation": ["logistic", "relu"],
    "solver": ["adam"],
    "learning_rate": ["constant"],
    "learning_rate_init": [0.001, 0.01],
    "alpha": [1e-4, 1e-3, 1e-2, 1e-1],
    "max_iter": [200],
    "early_stopping": [True, False],
}

RANDOM_FOREST_GRID = {
    "n_estimators": [100, 300, 500],
    "min_samples_split": [2, 0.05, 0.1],
    "min_samples_leaf": [1, 0.05, 0.1],
    "max_depth": [None, 3, 6],
}

XGB_GRID = {
    "n_estimators": [100, 300, 500],
    "max_depth": [None, 3, 6],
    "learning_rate": [0.01, 0.05, 0.1],
    "subsample": [0.15, 0.20, 0.25],
}

SVM_GRID = {
    "kernel": [k.lower() for k in SVM_KERNELS],
    "degree": [2, 3, 4],
    "C": [1.0, 10.0, 100],
}


MLREM_GRID = {
    "alpha": [0.05, 0.1, 0.5, 0.8],
    "max_beta": [40.0],
    "weight_threshold": [1e-3],
    "max_iterations": [300],
    "tolerance": [0.01],
}

BRNN_GRID = {
    "hidden_layer_sizes": [32, 64, 128],
    "activation": [
        ActivationFunctions.Logistic,
        ActivationFunctions.Tanh,
        ActivationFunctions.ReLU,
    ],
    "batch_size": [64, 128],
    "learning_rate_init": [0.01, 0.005, 0.001],
    "max_iter": [200, 300],
}



"""Feature Importance page for Helix.

This page provides options for analyzing feature importance using various methods:
- Global methods (Permutation, SHAP)
- Local methods (LIME, SHAP)
- Ensemble methods
- Fuzzy feature importance
"""

from multiprocessing import Process
from pathlib import Path

import pandas as pd
import streamlit as st

from helix.components.configuration import display_options
from helix.components.experiments import experiment_selector, model_selector
from helix.components.forms.forms_fi import fi_options_form
from helix.components.images.logos import sidebar_logo
from helix.components.logs import log_box
from helix.components.plots import fuzzy_feature_impacts, plot_box, plot_box_v2
from helix.feature_importance import feature_importance, fuzzy_interpretation
from helix.options.data import DataOptions
from helix.options.enums import (
    ExecutionStateKeys,
    FeatureImportanceStateKeys,
    FuzzyStateKeys,
    ViewExperimentKeys,
)
from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.file_paths import (
    data_options_path,
    execution_options_path,
    fi_options_path,
    fi_plot_dir,
    fuzzy_options_path,
    fuzzy_plot_dir,
    fuzzy_result_dir,
    helix_experiments_base_dir,
    log_dir,
    ml_model_dir,
    plot_options_path,
)
from helix.options.fuzzy import FuzzyOptions
from helix.options.plotting import PlottingOptions
from helix.services.configuration import (
    load_data_options,
    load_execution_options,
    load_plot_options,
    save_options,
)
from helix.services.data import TabularData, ingest_data
from helix.services.experiments import (
    delete_previous_fi_results,
    find_previous_fi_results,
    get_experiments,
)
from helix.services.logs import get_logs
from helix.services.ml_models import load_models_to_explain
from helix.utils.logging_utils import Logger, close_logger
from helix.utils.utils import cancel_pipeline, set_seed


def build_configuration() -> tuple[
    FuzzyOptions | None,
    FeatureImportanceOptions,
    ExecutionOptions,
    PlottingOptions,
    DataOptions,
    str,
    list,
]:
    """Build the configuration objects for the pipeline.

    Returns:
        tuple[
        FuzzyOptions | None,
        FeatureImportanceOptions,
        ExecutionOptions,
        PlottingOptions,
        DataOptions,
        str,
        list]:
        - The options for fuzzy,
        - The options for feature importance
        - The options for pipeline execution
        - The plotting options
        - The data options
        - The experiment name
        - The list of models to explain.
    """
    biofefi_base_dir = helix_experiments_base_dir()
    experiment_name = st.session_state[ExecutionStateKeys.ExperimentName]

    # Load plotting options
    path_to_plot_opts = plot_options_path(biofefi_base_dir / experiment_name)
    plotting_options = load_plot_options(path_to_plot_opts)

    # Load executuon options
    path_to_exec_opts = execution_options_path(biofefi_base_dir / experiment_name)
    exec_opt = load_execution_options(path_to_exec_opts)

    # Load data options
    path_to_data_opts = data_options_path(biofefi_base_dir / experiment_name)
    data_options = load_data_options(path_to_data_opts)

    # Set up fuzzy options
    fuzzy_opt = None
    if st.session_state.get(FuzzyStateKeys.FuzzyFeatureSelection, False):
        fuzzy_opt = FuzzyOptions(
            fuzzy_feature_selection=st.session_state[
                FuzzyStateKeys.FuzzyFeatureSelection
            ],
            number_fuzzy_features=st.session_state[
                FuzzyStateKeys.NumberOfFuzzyFeatures
            ],
            granular_features=st.session_state[FuzzyStateKeys.GranularFeatures],
            number_clusters=st.session_state[FuzzyStateKeys.NumberOfClusters],
            cluster_names=st.session_state.get(FuzzyStateKeys.ClusterNames, "").split(
                ", "
            ),
            number_rules=st.session_state[FuzzyStateKeys.NumberOfTopRules],
            save_fuzzy_set_plots=plotting_options.save_plots,
            fuzzy_log_dir=str(
                log_dir(
                    biofefi_base_dir
                    / st.session_state[ViewExperimentKeys.ExperimentName]
                )
                / "fuzzy"
            ),
        )

    # Set up feature importance options
    fi_opt = FeatureImportanceOptions(
        num_features_to_plot=st.session_state[
            FeatureImportanceStateKeys.NumberOfImportantFeatures
        ],
        permutation_importance_scoring=st.session_state[
            FeatureImportanceStateKeys.ScoringFunction
        ],
        permutation_importance_repeat=st.session_state[
            FeatureImportanceStateKeys.NumberOfRepetitions
        ],
        save_feature_importance_plots=plotting_options.save_plots,
        fi_log_dir=str(
            log_dir(
                biofefi_base_dir / st.session_state[ViewExperimentKeys.ExperimentName]
            )
            / "fi"
        ),
        save_feature_importance_options=st.session_state[
            FeatureImportanceStateKeys.SaveFeatureImportanceOptions
        ],
        save_feature_importance_results=st.session_state[
            FeatureImportanceStateKeys.SaveFeatureImportanceResults
        ],
        local_importance_methods=st.session_state[
            FeatureImportanceStateKeys.LocalImportanceFeatures
        ],
        feature_importance_ensemble=st.session_state[
            FeatureImportanceStateKeys.EnsembleMethods
        ],
        global_importance_methods=st.session_state[
            FeatureImportanceStateKeys.GlobalFeatureImportanceMethods
        ],
    )

    return (
        fuzzy_opt,
        fi_opt,
        exec_opt,
        plotting_options,
        data_options,
        experiment_name,
        st.session_state[FeatureImportanceStateKeys.ExplainModels],
    )


def pipeline(
    fuzzy_opts: FuzzyOptions,
    fi_opts: FeatureImportanceOptions,
    exec_opts: ExecutionOptions,
    plot_opts: PlottingOptions,
    experiment_name: str,
    explain_models: list,
    data: TabularData,
):
    """This function actually performs the steps of the pipeline. It can be wrapped
    in a process it doesn't block the UI.

    Args:
        fuzzy_opts (FuzzyOptions): Options for fuzzy feature importance.
        fi_opts (FeatureImportanceOptions): Options for feature importance.
        exec_opts (ExecutionOptions): Options for pipeline execution.
        plot_opts (PlottingOptions): Options for plotting.
        experiment_name (str): The experiment name.
        explain_models (list): The models to analyse.
        data (TabularData): The data that will be used in the pipeline.
    """
    biofefi_base_dir = helix_experiments_base_dir()
    seed = exec_opts.random_state
    set_seed(seed)

    # Models will already be trained before feature importance
    trained_models = load_models_to_explain(
        ml_model_dir(biofefi_base_dir / experiment_name), explain_models
    )

    fi_logger_instance = Logger(Path(fi_opts.fi_log_dir))
    fi_logger = fi_logger_instance.make_logger()

    # Load data options to get path to data
    path_to_data_opts = data_options_path(biofefi_base_dir / experiment_name)
    data_opts = load_data_options(path_to_data_opts)
    # Feature importance
    (
        gloabl_importance_results,
        local_importance_results,
        ensemble_results,
    ) = feature_importance.run(
        fi_opt=fi_opts,
        exec_opt=exec_opts,
        plot_opt=plot_opts,
        data_opt=data_opts,
        data=data,
        models=trained_models,
        data_path=Path(data_opts.data_path),
        logger=fi_logger,
    )
    # Close the fi logger
    close_logger(fi_logger_instance, fi_logger)

    # Fuzzy interpretation
    if fuzzy_opts is not None and fuzzy_opts.fuzzy_feature_selection:
        fuzzy_logger_instance = Logger(Path(fuzzy_opts.fuzzy_log_dir))
        fuzzy_logger = fuzzy_logger_instance.make_logger()
        fuzzy_interpretation.run(
            fuzzy_opt=fuzzy_opts,
            fi_opt=fi_opts,
            exec_opt=exec_opts,
            plot_opt=plot_opts,
            data=data,
            ensemble_results=ensemble_results,
            local_results=local_importance_results,
            logger=fuzzy_logger,
        )
        close_logger(fuzzy_logger_instance, fuzzy_logger)


def display_feature_importance_plots(experiment_dir: Path) -> None:
    """Display feature importance plots from the given experiment directory.

    Args:
        experiment_dir: Path to the experiment directory
    """
    fi_plots = fi_plot_dir(experiment_dir)
    if fi_plots.exists():
        mean_plots = [
            p
            for p in fi_plots.iterdir()
            if p.name.endswith("-all-folds-mean.png")  # mean global FI
            or p.name.startswith("local-")  # local plots
            or p.name.startswith("ensemble-")  # ensemble plots
        ]
        plot_box_v2(mean_plots, "Feature importance plots")


def display_fuzzy_plots(experiment_dir: Path) -> None:
    """Display fuzzy plots from the given experiment directory.

    Args:
        experiment_dir: Path to the experiment directory
    """
    fuzzy_plots = fuzzy_plot_dir(experiment_dir)
    fuzzy_impacts = fuzzy_result_dir(experiment_dir) / "top contextual rules.csv"
    if fuzzy_plots.exists():
        plot_box(fuzzy_plots, "Fuzzy plots")
    if fuzzy_impacts.exists():
        impacts_df = pd.read_csv(fuzzy_impacts)
        impacts_df.rename({"Unnamed: 0": ""}, axis=1, inplace=True)
        fuzzy_feature_impacts(impacts_df)


# Set page contents
st.set_page_config(
    page_title="Feature Importance",
    page_icon=sidebar_logo(),
)


st.header("Feature Importance")
# st.write(
#     """
#     This page provides options for exploring and customising feature importance and
#     interpretability methods in the trained machine learning models.
#     You can configure global and local feature importance techniques,
#     select ensemble approaches, and apply fuzzy feature importance.
#     Options include tuning scoring functions, setting data percentages
#     for SHAP analysis, and configuring rules for fuzzy synergy analysis
#     to gain deeper insights into model behaviour.
#     """
# )
st.write(
    """
    This page provides options for exploring and customising feature importance and
    interpretability methods in the trained machine learning models.
    You can configure global, local feature importance techniques and
    select ensemble approaches.
    """
)


choices = get_experiments()
experiment_name = experiment_selector(choices)
base_dir = helix_experiments_base_dir()


if experiment_name:

    previous_results_exist = find_previous_fi_results(
        helix_experiments_base_dir() / experiment_name
    )
    display_options(base_dir / experiment_name)

    path_to_exec_opts = execution_options_path(
        helix_experiments_base_dir() / experiment_name
    )
    exec_opt = load_execution_options(path_to_exec_opts)

    if previous_results_exist:
        st.warning("You have run feature importance in this experiment previously.")
        st.checkbox(
            "Would you like to rerun feature importance? This will overwrite the existing results.",
            value=True,
            key=FuzzyStateKeys.RerunFI,
        )
    else:
        st.session_state[FuzzyStateKeys.RerunFI] = True

    if st.session_state[FuzzyStateKeys.RerunFI]:

        st.session_state[ExecutionStateKeys.ExperimentName] = experiment_name
        st.session_state[ExecutionStateKeys.ProblemType] = exec_opt.problem_type

        model_dir = ml_model_dir(base_dir / experiment_name)
        if model_dir.exists():
            model_choices = list(
                filter(
                    lambda x: x.endswith(".pkl"), [x.name for x in model_dir.iterdir()]
                )
            )
        else:
            model_choices = []

        if len(model_choices) == 0:
            st.info(
                "You don't have any trained models in this experiment. "
                "Go to **Train Models** to create some models to evaulate."
            )
        elif st.toggle(
            "Explain all models",
            key=FeatureImportanceStateKeys.ExplainAllModels,
        ):
            st.session_state[FeatureImportanceStateKeys.ExplainModels] = model_choices
        else:
            model_selector(
                options=model_choices,
                gui_text="Select a model to explain",
                placeholder="Models to explain",
                key=FeatureImportanceStateKeys.ExplainModels,
            )

        if model_choices := st.session_state.get(
            FeatureImportanceStateKeys.ExplainModels
        ):
            fi_options_form()

            if st.button("Run Feature Importance", type="primary"):
                delete_previous_fi_results(base_dir / experiment_name)
                (
                    fuzzy_opts,
                    fi_opts,
                    exec_opts,
                    plot_opts,
                    data_opts,
                    exp_name,
                    models_to_explaion,
                ) = build_configuration()
                # Create the logger
                logger_instance = Logger()
                logger = logger_instance.make_logger()
                # Ingest the data
                data = ingest_data(exec_opts, data_opts, logger)
                # save FI options
                fi_options_file = fi_options_path(base_dir / experiment_name)
                save_options(fi_options_file, fi_opts)
                # save Fuzzy options if configured
                if fuzzy_opts is not None:
                    fuzzy_options_file = fuzzy_options_path(base_dir / experiment_name)
                    save_options(fuzzy_options_file, fuzzy_opts)

                process = Process(
                    target=pipeline,
                    args=(
                        fuzzy_opts,
                        fi_opts,
                        exec_opts,
                        plot_opts,
                        exp_name,
                        models_to_explaion,
                        data,
                    ),
                    daemon=True,
                )
                process.start()
                cancel_button = st.button(
                    "Cancel", on_click=cancel_pipeline, args=(process,)
                )
                with st.spinner(
                    "Feature Importance pipeline is running in the background. "
                    "Check the logs for progress."
                ):
                    # wait for the process to finish or be cancelled
                    process.join()
                try:
                    st.session_state[FeatureImportanceStateKeys.FILogBox] = get_logs(
                        log_dir(base_dir / experiment_name) / "fi"
                    )
                    st.session_state[FuzzyStateKeys.FuzzyLogBox] = get_logs(
                        log_dir(base_dir / experiment_name) / "fuzzy"
                    )
                    log_box(
                        box_title="Feature Importance Logs",
                        key=FeatureImportanceStateKeys.FILogBox,
                    )
                    # log_box(box_title="Fuzzy FI Logs", key=FuzzyStateKeys.FuzzyLogBox)
                except NotADirectoryError:
                    pass
                # Display plots
                display_feature_importance_plots(base_dir / experiment_name)
                display_fuzzy_plots(base_dir / experiment_name)

    else:
        st.success(
            "You have chosen not to rerun the feature importance experiments. "
            "You can proceed to see the experiment results."
        )



"""This module contains choices that will appear in the UI."""

from helix.options.enums import (
    DataSplitMethods,
    Normalisations,
    ProblemTypes,
    SvmKernels,
    TransformationsY,
)

SVM_KERNELS = [
    SvmKernels.RBF.upper(),  # appear as RBF, not Rbf
    SvmKernels.Linear.capitalize(),
    SvmKernels.Poly.capitalize(),
    SvmKernels.Sigmoid.capitalize(),
]
PROBLEM_TYPES = [
    ProblemTypes.Classification.capitalize(),
    ProblemTypes.Regression.capitalize(),
]
NORMALISATIONS = [
    Normalisations.Standardisation.capitalize(),
    Normalisations.MinMax.capitalize(),
    Normalisations.MeanCentering.capitalize(),
    Normalisations.MeanCenteringPoissonScaling.capitalize(),
    Normalisations.ParetoScaling.capitalize(),
    Normalisations.NoNormalisation.capitalize(),
]
TRANSFORMATIONS_Y = [
    TransformationsY.Log.capitalize(),
    TransformationsY.Sqrt.capitalize(),
    TransformationsY.MinMaxNormalisation.capitalize(),
    TransformationsY.StandardisationNormalisation.capitalize(),
    TransformationsY.NoTransformation.capitalize(),
]

PLOT_FONT_FAMILIES = ["serif", "sans-serif", "cursive", "fantasy", "monospace"]
DATA_SPLITS = [
    DataSplitMethods.Holdout.capitalize(),
    DataSplitMethods.KFold.capitalize(),
]



import streamlit as st

from helix.components.images.logos import header_logo, sidebar_logo

st.set_page_config(
    page_title="Helix",
    page_icon=sidebar_logo(),
)
header_logo()
sidebar_logo()

st.write("# Welcome")
st.write(
    """
    **Helix** lets you **rapidly** develop machine learning models of many kinds, and evaluate their performance down to a **feature-by-feature** level.

    You can create models to solve either **classification** problems (e.g. is this image a cat 🐱 or a dog 🐶?)
    or **regression** problems (e.g. what will be the price of gold 🏅 tomorrow 📈?).

    Your models can then be evaluated by general measures, such as **accuracy**, and by individual feature metrics,
    such as **SHAP**.

    ### Using Helix

    To create a **new experiment** ⚗️, go to the sidebar on the **left** and click **"New Experiment"**.

    To preprocess your data, go to the sidebar on the **left** and click **"Data Preprocessing"**.

    To visualise your data as part of your exploratory data analysis, go to the sidebar on the **left** and click **"Data Visualisation"**.

    To train new machine learning models 🏋️, go to the sidebar on the **left** and click **"Train Models"**.

    To run a feature importance analysis 📊, go to the sidebar on the **left** and click **"Feature Importance"**.

    To view your previous experiments 📈, go to the sidebar on the **left** and click **"View Experiments"**.
"""
)
st.markdown(
    '<span style="color:red; font-size:1em;"><i>You are using Helix 1.1.0</i></span>',
    unsafe_allow_html=True,
)



import os
import warnings
from pathlib import Path

import numpy as np
from sklearn.preprocessing import OneHotEncoder

from helix.options.enums import Metrics, ModelNames, ProblemTypes
from helix.options.execution import ExecutionOptions
from helix.options.ml import MachineLearningOptions
from helix.options.plotting import PlottingOptions
from helix.services.data import TabularData
from helix.services.plotting import (
    plot_auc_roc,
    plot_beta_coefficients,
    plot_confusion_matrix,
    plot_scatter,
)
from helix.utils.logging_utils import Logger
from helix.utils.plotting import close_figure

warnings.filterwarnings("ignore", message="X has feature names")


def _get_metric_for_problem_type(problem_type: ProblemTypes) -> str:
    """Get the appropriate metric based on the problem type.

    Args:
        problem_type (ProblemTypes): The type of problem (Regression or Classification)

    Returns:
        str: The metric to use
    """
    return Metrics.R2 if problem_type == ProblemTypes.Regression else Metrics.ROC_AUC


def _find_closest_bootstrap_index(
    ml_metric_results: dict,
    ml_metric_results_stats: dict,
    model_name: str,
    metric: str,
) -> int:
    """Find the bootstrap index closest to the mean metric value.

    Args:
        ml_metric_results (dict): The machine learning results
        ml_metric_results_stats (dict): The statistics for the ML metrics
        model_name (str): Name of the model
        metric (str): Metric to use for comparison

    Returns:
        int: Index of the bootstrap closest to the mean
    """
    mean_metric_test = ml_metric_results_stats[model_name]["test"][metric]["mean"]
    closest_index = 0
    min_diff = float("inf")

    for i, bootstrap in enumerate(ml_metric_results[model_name]):
        metric_test_value = bootstrap[metric]["test"]["value"]
        current_diff = abs(metric_test_value - mean_metric_test)
        if current_diff < min_diff:
            min_diff = current_diff
            closest_index = i

    return closest_index


def _save_regression_plots(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    metric_results: dict,
    split_type: str,
    dependent_variable: str,
    model_name: str,
    directory: Path,
    plot_opts: PlottingOptions,
    logger: Logger,
) -> None:
    """Save regression plots for a specific split.

    Args:
        y_true (np.ndarray): True values
        y_pred (np.ndarray): Predicted values
        metric_results (dict): Results for the metrics
        split_type (str): Type of split (Train/Test)
        dependent_variable (str): Name of dependent variable
        model_name (str): Name of the model
        directory (Path): Directory to save plots
        bootstrap_index (int): Index of the bootstrap
        plot_opts (PlottingOptions): Plot options
    """
    try:

        if model_name.lower() == "linear model":
            display_model_name = "Linear Regression"
        else:
            display_model_name = " ".join(
                word.capitalize() for word in model_name.split()
            )

        figure_title = f"Prediction Error for {display_model_name} - {split_type}"
        plot_opts.plot_title = figure_title

        plot_opts.width, plot_opts.height = (
            10,
            8,
        )  # this is to match the previous default size for parity plots

        fig = plot_scatter(
            y=y_true,
            yp=y_pred,
            r2=metric_results["R2"][split_type.lower()],
            dependent_variable=dependent_variable,
            plot_opts=plot_opts,
        )
        save_path = directory / f"{model_name}-{split_type.lower()}-scatter.png"
        fig.savefig(save_path, dpi=plot_opts.dpi, bbox_inches="tight")
        close_figure(fig)
    except Exception as e:
        logger.error(f"Error creating scatter plot: {str(e)}")


def _save_coefficient_plot(
    model,
    feature_names: list,
    plot_opts: PlottingOptions,
    model_name: str,
    dependent_variable: str,
    directory: Path,
    problem_type: ProblemTypes,
    logger: Logger,
) -> None:
    """Save coefficient plot for linear regression models.

    Args:
        model: The trained model
        feature_names (list): List of feature names
        plot_opts (PlottingOptions): Plot options
        model_name (str): Name of the model
        dependent_variable (str): Name of dependent variable
        directory (Path): Directory to save plots
        bootstrap_index (int): Index of the bootstrap
        problem_type (ProblemTypes): Type of problem (Classification or Regression)
    """
    if hasattr(model, "coef_"):
        try:
            coefficients = model.coef_
            if len(coefficients.shape) > 1 and coefficients.shape[0] == 1:
                # Handle case where coefficients are in shape (1, n_features)
                coefficients = coefficients.reshape(-1)

            # Ensure feature_names matches coefficient length
            if len(coefficients) != len(feature_names):
                raise ValueError(
                    f"Number of coefficients ({len(coefficients)}) does not match "
                    f"number of feature names ({len(feature_names)})"
                )

            fig = plot_beta_coefficients(
                coefficients=coefficients,
                feature_names=feature_names,
                plot_opts=plot_opts,
                model_name=model_name,
                dependent_variable=dependent_variable,
                is_classification=(problem_type == ProblemTypes.Classification),
            )
            save_path = directory / f"{model_name}-coefficients.png"
            fig.savefig(save_path, dpi=plot_opts.dpi, bbox_inches="tight")
            close_figure(fig)
        except Exception as e:
            logger.error(f"Error creating coefficient plot: {str(e)}")


def _save_classification_plots(
    y_true: np.ndarray,
    y_pred_proba: np.ndarray,
    split_type: str,
    model_name: str,
    directory: Path,
    plot_opts: PlottingOptions,
    logger: Logger,
) -> None:
    """Save classification plots for a specific split.

    Args:
        y_true (np.ndarray): True values
        y_pred_proba (np.ndarray): Predicted probabilities
        model: The trained model
        split_type (str): Type of split (Train/Test)
        model_name (str): Name of the model
        directory (Path): Directory to save plots
        plot_opts (PlottingOptions): Plot options
        logger (Logger): The logger instance
        feature_names (list | None, optional): List of feature names. Defaults to None.
    """
    try:
        encoder = OneHotEncoder()
        encoder.fit(y_true.reshape(-1, 1))
        y_true_labels = encoder.transform(y_true.reshape(-1, 1)).toarray()

        plot_auc_roc(
            y_classes_labels=y_true_labels,
            y_score_probs=y_pred_proba,
            set_name=split_type,
            model_name=model_name,
            directory=directory,
            plot_opts=plot_opts,
        )
    except Exception as e:
        logger.error(f"Error plotting ROC curve: {str(e)}")

    try:
        # Get predictions for confusion matrix
        y_pred = (
            np.argmax(y_pred_proba, axis=1) if y_pred_proba.ndim > 1 else y_pred_proba
        )

        # Plot confusion matrix
        plot_confusion_matrix(
            y_true=y_true,  # True labels
            y_pred=y_pred,  # Predicted labels
            set_name=split_type,
            model_name=model_name,
            directory=directory,
            plot_opts=plot_opts,
        )
    except Exception as e:
        logger.error(f"Error plotting confusion matrix: {str(e)}")


def save_actual_pred_plots(
    data: TabularData,
    ml_results: dict,
    logger: Logger,
    ml_metric_results: dict,
    ml_metric_results_stats: dict,
    exec_opts: ExecutionOptions,
    plot_opts: PlottingOptions | None = None,
    ml_opts: MachineLearningOptions | None = None,
    trained_models: dict | None = None,
) -> None:
    """Save actual vs prediction plots for classification and regression.

    Args:
        data (DataBuilder): The data
        ml_results (dict): The machine learning results
        logger (Logger): The logger
        ml_metric_results (dict): The results for the ML metrics
        ml_metric_results_stats (dict): The statistics for the ML metrics
        n_bootstraps (int): The number of bootstraps or k-folds
        exec_opts (ExecutionOptions): The execution options
        plot_opts (PlottingOptions | None, optional): The plot options. Defaults to None
        ml_opts (MachineLearningOptions | None, optional): The ML options. Defaults to None
        trained_models (dict | None, optional): The ML models. Defaults to None
    """
    metric = _get_metric_for_problem_type(exec_opts.problem_type)
    # Create results directory
    directory = Path(ml_opts.ml_plot_dir)
    os.makedirs(directory, exist_ok=True)
    # Convert data to numpy arrays
    y_test = [np.array(df) for df in data.y_test]
    y_train = [np.array(df) for df in data.y_train]
    # Process each model
    for model_name, model_options in ml_opts.model_types.items():
        if not model_options["use"]:
            continue
        logger.info(f"Saving actual vs prediction plots of {model_name}...")
        # Find best bootstrap index
        closest_index = _find_closest_bootstrap_index(
            ml_metric_results, ml_metric_results_stats, model_name, metric
        )
        # Get predictions for best bootstrap
        y_pred_test = ml_results[closest_index][model_name]["y_pred_test"]
        y_pred_train = ml_results[closest_index][model_name]["y_pred_train"]
        if exec_opts.problem_type == ProblemTypes.Regression:
            # Save regression plots
            _save_regression_plots(
                y_true=y_test[closest_index],
                y_pred=y_pred_test,
                metric_results=ml_metric_results[model_name][closest_index],
                split_type="Test",
                dependent_variable=exec_opts.dependent_variable,
                model_name=model_name,
                directory=directory,
                plot_opts=plot_opts,
                logger=logger,
            )
            _save_regression_plots(
                y_true=y_train[closest_index],
                y_pred=y_pred_train,
                metric_results=ml_metric_results[model_name][closest_index],
                split_type="Train",
                dependent_variable=exec_opts.dependent_variable,
                model_name=model_name,
                directory=directory,
                plot_opts=plot_opts,
                logger=logger,
            )
            # Save coefficient plots for linear models
            if model_name in [
                ModelNames.LinearModel.value,
                ModelNames.MLREM.value,
                ModelNames.Lasso,
                ModelNames.Ridge,
                ModelNames.ElasticNet,
            ]:
                _save_coefficient_plot(
                    model=trained_models[model_name][closest_index],
                    feature_names=data.X_train[closest_index].columns.tolist(),
                    plot_opts=plot_opts,
                    model_name=model_name,
                    dependent_variable=exec_opts.dependent_variable,
                    directory=directory,
                    problem_type=exec_opts.problem_type,
                    logger=logger,
                )
        else:
            # Save classification plots
            y_pred_test_proba = ml_results[closest_index][model_name][
                "y_pred_test_proba"
            ]
            y_pred_train_proba = ml_results[closest_index][model_name][
                "y_pred_train_proba"
            ]
            _save_classification_plots(
                y_true=y_test[closest_index],
                y_pred_proba=y_pred_test_proba,
                split_type="Test",
                model_name=model_name,
                directory=directory,
                plot_opts=plot_opts,
                logger=logger,
            )
            _save_classification_plots(
                y_true=y_train[closest_index],
                y_pred_proba=y_pred_train_proba,
                split_type="Train",
                model_name=model_name,
                directory=directory,
                plot_opts=plot_opts,
                logger=logger,
            )
            # Save coefficient plots for linear classification models
            if model_name == "linear model" and hasattr(
                trained_models[model_name][closest_index], "coef_"
            ):
                _save_coefficient_plot(
                    model=trained_models[model_name][closest_index],
                    feature_names=data.X_train[closest_index].columns.tolist(),
                    plot_opts=plot_opts,
                    model_name=model_name,
                    dependent_variable=exec_opts.dependent_variable,
                    directory=directory,
                    problem_type=exec_opts.problem_type,
                    logger=logger,
                )



from helix.feature_importance.fuzzy import Fuzzy
from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.fuzzy import FuzzyOptions
from helix.options.plotting import PlottingOptions


def run(
    fuzzy_opt: FuzzyOptions,
    fi_opt: FeatureImportanceOptions,
    exec_opt: ExecutionOptions,
    plot_opt: PlottingOptions,
    data,
    ensemble_results,
    local_results,
    logger,
):

    # Interpret the feature synergy importance
    fuzzy = Fuzzy(
        fuzzy_opt=fuzzy_opt,
        fi_opt=fi_opt,
        exec_opt=exec_opt,
        plot_opt=plot_opt,
        logger=logger,
    )
    fuzzy_rules = fuzzy.interpret(ensemble_results, local_results, data)

    return fuzzy_rules
